/*
	Copyright (C) 2006,2007 John Anderson

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef mackie_controls_h
#define mackie_controls_h

#include <map>
#include <vector>
#include <string>

#include "mackie_control_exception.h"

namespace Mackie
{

class Control;

/**
	This is a loose group of controls, eg cursor buttons,
	transport buttons, functions buttons etc.
*/
class Group
{
public:
	Group( const std::string & name )
	: _name( name )
	{
	}
	
	virtual ~Group() {}
	
	virtual bool is_strip() const
	{
		return false;
	}
	
	virtual void add( Control & control );
	
	const std::string & name() const
	{
		return _name;
	}

private:
	std::string _name;
	std::vector<Control*> controls;
};

class Button;
class Pot;
class Fader;

/**
	This is the set of controls that make up a strip.
*/
class Strip : public Group
{
public:
	Strip( const std::string & name )
	: Group( name )
	{
	}
	
	virtual bool is_strip() const
	{
		return true;
	}
	
	virtual void add( Control & control );
	
	Button & solo();
	Button & recenable();
	Button & mute();
	Button & select();
	Button & vselect();
	Button & fader_touch();
	Pot & vpot();
	Fader & gain();
	
private:
	Button * _solo;
	Button * _recenable;
	Button * _mute;
	Button * _select;
	Button * _vselect;
	Button * _fader_touch;
	Pot * _vpot;
	Fader * _gain;
};

class Led;

/**
	The base class for controls on the surface. They deliberately
	don't know the midi protocol for updating them.
*/
class Control
{
public:
	Control( int id, int ordinal, std::string name, Group & group )
	: _id( id ), _ordinal( ordinal ), _name( name ), _group( group )
	{
		_group.add( *this );
	}
	
	virtual ~Control() {}
	
	virtual Led & led() const
	{
		if ( _led == 0 )
		{
			throw MackieControlException( "_led is null in Control" );
		}
		return *_led;
	}

	/// The midi id of the control
	int id() const
	{
		return _id;
	}
	
	/// The 1-based number of the control
	int ordinal() const
	{
		return _ordinal;
	}
	
	const std::string & name() const
	{
		return _name;
	}
	
	const Group & group() const
	{
		return _group;
	}
	
private:
	int _id;
	int _ordinal;
	std::string _name;
	Group & _group;
	Led * _led;
};

std::ostream & operator << ( std::ostream & os, const Control & control );

class Fader : public Control
{
public:
	Fader( int id, int ordinal, std::string name, Group & group )
	: Control( id, ordinal, name, group )
	{
	}
};

class Button : public Control
{
public:
	Button( int id, int ordinal, std::string name, Group & group )
	: Control( id, ordinal, name, group )
	{
	}
};

class Pot : public Control
{
public:
	Pot( int id, int ordinal, std::string name, Group & group )
	: Control( id, ordinal, name, group )
	{
	}
};

class Led : public Control
{
public:
	Led( int id, int ordinal, std::string name, Group & group )
	: Control( id, ordinal, name, group )
	{
	}
};

class LedRing : public Led
{
public:
	LedRing( int id, int ordinal, std::string name, Group & group )
	: Led( id, ordinal, name, group )
	{
	}
};

/**
	This represents an entire control surface, made up of Groups,
	Strips and Controls. There are several collections for
	ease of addressing in different ways, but only one collection
	has definitive ownership.

	TODO maybe make Group inherit from Control, for ease of ownership.
*/
class Surface
{
public:
	Surface( uint32_t max_strips );
	
	~Surface();

	typedef std::vector<Control*> Controls;
	
	/// This collection has ownership of all the controls
	Controls controls;

	/**
		These are alternative addressing schemes
		They use maps because the indices aren't always
		0-based.
	*/
	std::map<int,Control*> faders;
	std::map<int,Control*> pots;
	std::map<int,Control*> buttons;
	std::map<int,Control*> leds;
	std::vector<Strip*> strips;

	/// This collection owns the groups
	typedef std::map<std::string,Group*> Groups;
	Groups groups;

	uint32_t max_strips() const
	{
		return _max_strips;
	}
	
protected:
	void init_controls();

private:
	uint32_t _max_strips;
};

}

#endif
