/*
	Copyright (C) 2006,2007 Paul Davis, John Anderson

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	$Id: mackie_control_protocol.cc 1059 2006-11-02 12:27:49Z paul $
*/

#include <iostream>
#include <algorithm>
#include <cmath>

#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <float.h>
#include <sys/time.h>
#include <errno.h>

#include <boost/shared_array.hpp>

#include <pbd/pthread_utils.h>

#include <ardour/route.h>
#include <ardour/audio_track.h>
#include <ardour/session.h>
#include <ardour/location.h>
#include <ardour/dB.h>
#include <ardour/panner.h>
#include <midi++/types.h>
#include <midi++/port.h>
#include <midi++/manager.h>
#include <midi++/port_request.h>

#include "mackie_control_protocol.h"
#include "midi_byte_array.h"
#include "mackie_control_exception.h"
#include "route_signal.h"

using namespace ARDOUR;
using namespace std;
using namespace sigc;
using namespace PBD;
using namespace Mackie;

#include "i18n.h"

MidiByteArray MackieControlProtocol::mackie_sysex_hdr ( 5, MIDI::sysex, 0x0, 0x0, 0x66, 0x10 );

// Copied from tranzport_control_protocol.cc
static inline double 
gain_to_slider_position (ARDOUR::gain_t g)
{
	if (g == 0) return 0;
	return pow((6.0*log(g)/log(2.0)+192.0)/198.0, 8.0);
}

/*
	Copied from tranzport_control_protocol.cc
	TODO this seems to return slightly wrong values, namely
	with the UI slider at max, we get a 0.99something value.
*/
static inline ARDOUR::gain_t 
slider_position_to_gain (double pos)
{
	/* XXX Marcus writes: this doesn't seem right to me. but i don't have a better answer ... */
	if (pos == 0.0) return 0;
	return pow (2.0,(sqrt(sqrt(sqrt(pos)))*198.0-192.0)/6.0);
}

/**
	The first byte of a midi message from the surface
	will contain one of these, sometimes bitmasked
	with the control id
*/
enum midi_types {
	midi_fader_id = 0xe0
	, midi_button_id = 0x90
	, midi_pot_id = 0xb0
};

/**
	The LED rings have these modes.
*/
enum midi_pot_mode {
	midi_pot_mode_dot = 0
	, midi_pot_mode_boost_cut = 1
	, midi_pot_mode_wrap = 2
	, midi_pot_mode_spread = 3
};

// This should be moved to midi++ at some point
ostream & operator << ( ostream & os, const MIDI::Port & port )
{
	os << "device: " << port.device ();
	os << "; ";
	os << "name: " << port.name ();
	os << "; ";
	os << "type: " << port.type ();
	os << "; ";
	os << "mode: " << port.mode ();
	os << "; ";
	os << "ok: " << port.ok ();
	os << "; ";
	os << "number: " << port.number ();
	os << "; ";
	return os;
}

MackieControlProtocol::MackieControlProtocol (Session& session)
	: ControlProtocol  (session, X_("Mackie"))
	, _surface( 8 )
{
	MIDI::Manager* mm = MIDI::Manager::instance();

	/* 
		TODO This will need to take into account the possibility that MCUXTs are
		plugged into other midi ports.
	*/
	
	_port = mm->port (X_("mackie"));

	if (_port == 0) {
		error << _("no MIDI port named \"control\" exists - generic MIDI control disabled") << endmsg;
		throw failed_constructor();
	}
	
	// set up the route table
	set_route_table_size( surface().max_strips() );
	
	// do the initial bank switch to connect signals
	switch_banks( 0 );
	
	// TODO update global buttons and displays
}

uint32_t MackieControlProtocol::index_from_remote_id( uint32_t remote_id )
{
	return remote_id - _current_initial_bank;
}

void MackieControlProtocol::switch_banks( uint32_t initial )
{
	_current_initial_bank = initial;
	
	// first clear the signals from old routes
	// taken care of by the RouteSignal destructors
	for( RouteSignals::iterator it = route_signals.begin(); it != route_signals.end(); ++it )
	{
		delete *it;
	}
	route_signals.clear();
	
	// now set the signals for new routes
	if ( initial <= session->nroutes() )
	{
		uint32_t i = 0;
		for (; i < route_table.size(); ++i )
		{
			boost::shared_ptr<Route> route = session->route_by_remote_id ( i );
			route_table[i] = route;
			if ( route != 0 )
			{
				route_signals.push_back(
					new RouteSignal( *route, *this, *surface().strips[i] )
				);
			}
		}
	}
	
	// TODO update all strips
}

MackieControlProtocol::~MackieControlProtocol ()
{
	set_active (false);
}

Control & MackieControlProtocol::lookup_control( int midi_type, const MidiByteArray & bytes )
{
	Control * control = 0;
	int midi_id = -1;
	switch( midi_type )
	{
		// fader
		case midi_fader_id:
			midi_id = bytes[0] & 0x0f;
			control = surface().faders[midi_id];
			if ( control == 0 )
			{
				ostringstream os;
				os << "control for fader" << midi_id << " is null";
				throw MackieControlException( os.str() );
			}
			break;
			
		// button
		case midi_button_id:
			midi_id = bytes[1];
			control = surface().buttons[midi_id];
			if ( control == 0 )
			{
				ostringstream os;
				os << "control for button" << midi_id << " is null";
				throw MackieControlException( os.str() );
			}
			break;
			
		// pot (jog wheel, external control)
		case midi_pot_id:
			midi_id = bytes[1] & 0x1f;
			control = surface().pots[midi_id];
			if ( control == 0 )
			{
				ostringstream os;
				os << "control for button" << midi_id << " is null";
				throw MackieControlException( os.str() );
			}
			break;
		
		default:
			ostringstream os;
			os << "Cannot find control for " << bytes;
			throw MackieControlException( os.str() );
	}
	return *control;
}

int
MackieControlProtocol::set_active (bool yn)
{
	if (yn != _active) {

		if (yn) {

			if (open ()) {
				return -1;
			}
			
			if (pthread_create_and_store (X_("mackie monitor"), &thread, 0, _monitor_work, this) == 0) {
				_active = true;
			} else {
				return -1;
			}

		} else {
			pthread_cancel_one (thread);
			close ();
			_active = false;
		} 
	}

	return 0;
}

bool MackieControlProtocol::handle_button( Control & control )
{
	bool state = false;
	uint32_t index = control.ordinal() - 1;

	if ( control.name() == "recenable" )
	{
		state = !route_get_rec_enable( index );
		route_set_rec_enable( index, state );
	}
	else if ( control.name() == "mute" )
	{
		state = !route_get_muted( index );
		route_set_muted( index, state );
	}
	else if ( control.name() == "solo" )
	{
		state = !route_get_soloed( index );
		route_set_soloed( index, state );
	}
	else if ( control.name() == "select" )
	{
		// TODO make the track selected. Whatever that means.
	}
	else if ( control.name() == "vselect" )
	{
		// TODO could be used to select different things to apply the pot to?
	}
	return state;
}

MIDI::byte calculate_pot_value( midi_pot_mode mode, float pos )
{
	MIDI::byte retval = 0x40; //0b01000000
	retval |= ( mode << 4 );
	retval += ( int(pos * 10.0) + 1 ) & 0x0f; // 0b00001111
	
	// TODO there's also a bit for center led
	return retval;
}

void MackieControlProtocol::handle_midi_any (MIDI::Parser & parser, MIDI::byte * raw_bytes, size_t count )
{
	MidiByteArray bytes( count, raw_bytes );
	try
	{
		// ignore sysex messages
		if ( bytes[0] == MIDI::sysex ) return;
			
		MIDI::byte midi_type = bytes[0] & 0xf0; //0b11110000
		Control & control = lookup_control( midi_type, bytes );
		uint32_t index = control.ordinal() - 1;
		
		cout << bytes << ' ' << control << endl;
		
		boost::shared_ptr<Route> route;
		if ( control.group().is_strip() )
		{
			route = route_table[index];
		}
		
		switch ( midi_type )
		{
			// fader
			case midi_fader_id:
				// find the route in the route table for the id
				// if the route isn't available, skip it
				// at which point the fader should just reset itself
				if ( route != 0 )
				{
					// for a BCF2000, max is 7f for high-order byte and 0x70 for low-order byte
					// According to the docs, these should both be 0x7f.
					int midi_pos = ( bytes[2] << 7 ) + bytes[1];
					double pos = double(midi_pos) / double(0x3fff);
					route_set_gain( index, slider_position_to_gain( pos ) );
					// make sure result is echoed back to surface
					write ( bytes );
				}
				break;
				
			// button
			case midi_button_id:
				if ( control.group().is_strip() )
				{
					if ( route != 0 )
					{
						if ( bytes[2] == 0x7f )
						{
							// this is a button press, so set the light
							// to whatever the ui-control says
							// it should be
							bytes[2] = handle_button( control ) ? 0x7f : 0x00;
							write ( bytes );
						}
						// otherwise it's a button-release, so leave the light alone
					}
					else
					{
						// no route so always switch the light off
						bytes[2] = 0x00;
						write ( bytes );
					}
				}
				else
				{
					// TODO must be one of the global buttons
				}
				break;
				
			// pot (jog wheel, external control)
			case midi_pot_id:
				if ( route != 0 )
				{
					if ( route->panner().size() == 1 )
					{
						// assume pan for now
						
						// bytes[2] & 0b01000000 (0x40) give sign
						int sign = ( bytes[2] & 0x40 ) == 0 ? 1 : -1; 
						// bytes[2] & 0b00111111 (0x3f) gives delta
						float delta = ( float( bytes[2] & 0x3f ) / float( 0x3f ) ) * sign;

						float xpos;
						route->panner()[0]->get_effective_position (xpos);
						
						// calculate new value, and trim
						// TODO do an exact calc for 0.50
						// and relate to center led, below
						xpos += delta;
						if ( xpos > 1.0 )
							xpos = 1.0;
						else if ( xpos < 0.0 )
							xpos = 0.0;
						
						route->panner()[0]->set_position( xpos );
						
						// now calculate outgoing pan value
						bytes[1] = calculate_pot_value( midi_pot_mode_dot, xpos );
						
						write( bytes );
					}
				}
				else
				{
					// TODO jog wheel or external
				}
				break;
		}
	}
	catch( MackieControlException & e )
	{
		cout << bytes << ' ' << e.what() << endl;
	}
}

/*
	This really only handles the control surface initialisation. It's
	kicked off by MackieControlProtocol::open.
*/
void MackieControlProtocol::handle_midi_sysex (MIDI::Parser & parser, MIDI::byte * raw_bytes, size_t count )
{
	try
	{
		cout << "MackieControlProtocol::handle_midi_sysex" << endl;
		MidiByteArray bytes( count, raw_bytes );
		cout << "sysex received: " << bytes << endl;
		
		MidiByteArray buf;
		back_insert_iterator<MidiByteArray> buf_insert(buf);
		
		switch ( sysex_state )
		{
			case expecting_host_connection_query:
				// decode host connection query
				if ( bytes.size() != 18 )
				{
					ostringstream os;
					os << "expecting 18 bytes, read " << bytes << " from " << port().name();
					throw MackieControlException( os.str() );
				}
				
				if ( bytes[5] != 1 )
				{
					ostringstream os;
					os << _("Mackie: incorrect status for Host Connection Query message") << (int)bytes[5];
					throw MackieControlException( os.str() );
				}
				
				// build and send host connection reply
				buf << mackie_sysex_hdr << 0x02;
				copy( bytes.begin() + 6, bytes.begin() + 6 + 11, buf_insert );
				buf << MIDI::eox;
				
				sysex_state = host_connection_confirmation;
				write( buf );
				break;
			
			case host_connection_confirmation:
				// decode host connection confirmation
				if ( bytes.size() != 14 )
				{
					ostringstream os;
					os << "expecting 14 bytes, read " << bytes << " from " << port().name();
					throw MackieControlException( os.str() );
				}
				
				if ( bytes[5] != 3 )
				{
					ostringstream os;
					os << _("incorrect status for Host Connection Confirmation message") << (int)bytes[5];
					throw MackieControlException( os.str() );
				}
				
				// allow other messages to flow
				port().input()->any.connect( ( mem_fun (*this, &MackieControlProtocol::handle_midi_any) ) );

				break;
			
			case initialisation_error:
				// do nothing
				break;
			
			default:
				sysex_state = initialisation_error;
				ostringstream os;
				os << "Unknown state " << sysex_state;
				throw MackieControlException( os.str() );
		}
	}
	catch( exception & e )
	{
		sysex_state = initialisation_error;
		error << e.what() << endmsg;
	}
}

/**
	This doesn't work because the reads of taken care of by the port
	parser.

	TODO revisit this once the port isn't mmc= in ardour.rc
*/
MidiByteArray MackieControlProtocol::read( int count_expecting )
{
	int count = 0;
	const int max_buf_size = 512;
	MIDI::byte raw_response[max_buf_size];
	MidiByteArray retval;
	
	if ( count == -1 )
	{
		ostringstream os;
		os << _("Mackie: couldn't read from port ") << port().name();
		throw MackieControlException( os.str() );
	}
	
	retval.copy( count, raw_response );
	cout << "read: " << retval << endl;
	
	if ( count_expecting != -1 && count != count_expecting )
	{
		ostringstream os;
		os << _("Mackie:") << "expecting " << count_expecting << " bytes but received " << count << " on " << port().name();
		throw MackieControlException( os.str() );
	}
	
	return retval;
}

void MackieControlProtocol::write( MidiByteArray & mba )
{
	cout << "writing: " << mba << endl;
	int count = port().write( mba.bytes().get(), mba.size() );
	if ( count != (int)mba.size() )
	{
		ostringstream os;
		os << _("Mackie: couldn't write to port ") << port().name();
		throw MackieControlException( os.str() );
	}
}

int
MackieControlProtocol::open ()
{
	try
	{
		/*
			connect to sysex messages. Connection to other
			messages done in handle_midi_sysex once
			the initialisation messages are complete
		*/
		port().input()->sysex.connect( ( mem_fun (*this, &MackieControlProtocol::handle_midi_sysex) ) );
		
		// set expecting state. Response will be taken care of by sysex handler.
		sysex_state = expecting_host_connection_query;
		
		// build and send device query
		MidiByteArray buf;
		buf << mackie_sysex_hdr << 0x0 << MIDI::eox;
		write( buf );
	}
	catch( exception & e )
	{
		error << e.what() << endmsg;
		return -1;
	}

	return 0;
}

void*
MackieControlProtocol::_monitor_work (void* arg)
{
	return static_cast<MackieControlProtocol*>(arg)->monitor_work ();
}

XMLNode&
MackieControlProtocol::get_state () 
{
	XMLNode* node = new XMLNode (X_("Protocol"));
	node->add_property (X_("name"), _name);
	return *node;
}

int
MackieControlProtocol::set_state (const XMLNode& node)
{
	return 0;
}

int
MackieControlProtocol::close ()
{
	error << "TODO: implement MackieControlProtocol::close" << endmsg;
	return 0;
}
	
void*
MackieControlProtocol::monitor_work ()
{
	struct sched_param rtparam;
	int err;

	PBD::ThreadCreated (pthread_self(), X_("Mackie"));

	memset (&rtparam, 0, sizeof (rtparam));
	rtparam.sched_priority = 3; /* XXX should be relative to audio (JACK) thread */
	
	if ((err = pthread_setschedparam (pthread_self(), SCHED_FIFO, &rtparam)) != 0) {
		// do we care? not particularly.
		PBD::info << string_compose (_("%1: thread not running with realtime scheduling (%2)"), name(), strerror (errno)) << endmsg;
	} 

	pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, 0);
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, 0);

	while (true) {
		sleep(1);
		
		// This is where the surface will be updated with the current
		// state according to the UI
		
		// Remember to check 'touched' status of faders
		
		// what does pthread_testcancel() do?
	}

	return (void*) 0;
}

Strip & MackieControlProtocol::strip_from_route( Route * route )
{
	if ( route == 0 )
	{
		throw MackieControlException( "route is null in MackieControlProtocol::strip_from_remote_id" );
	}
	uint32_t index = index_from_remote_id( route->remote_control_id() );
	return *surface().strips[index];
}

void MackieControlProtocol::notify_solo_changed( void *, ARDOUR::Route * route )
{
	try
	{
		Button & button = strip_from_route( route ).solo();
		MidiByteArray bytes( 3
			, midi_button_id
			, button.id()
			, route->soloed() ? 0x7f : 0
		);
		write( bytes );
	}
	catch( exception & e )
	{
		cout << e.what() << endl;
	}
}

void MackieControlProtocol::notify_mute_changed( void *, ARDOUR::Route * route )
{
	try
	{
		Button & button = strip_from_route( route ).mute();
		MidiByteArray bytes( 3, midi_button_id, button.id(), route->muted() ? 0x7f : 0 );
		write( bytes );
	}
	catch( exception & e )
	{
		cout << e.what() << endl;
	}
}

void MackieControlProtocol::notify_record_enable_changed( void *, ARDOUR::Route * route )
{
	try
	{
		Button & button = strip_from_route( route ).recenable();
		MidiByteArray bytes( 3, midi_button_id, button.id(), route->record_enabled() ? 0x7f : 0 );
		write( bytes );
	}
	catch( exception & e )
	{
		cout << e.what() << endl;
	}
}

// TODO make sure bytes are written only if fader is not touched
void MackieControlProtocol::notify_gain_changed( void *, ARDOUR::Route * route )
{
	try
	{
		Fader & fader = strip_from_route( route ).gain();
		int pos = int( 0x3fff * gain_to_slider_position( route->gain() ) );
		
		MidiByteArray bytes( 3
			, midi_fader_id | fader.id()
			// lower-order bits
			, pos & 0x7f
			// higher-order bits
			, ( pos >> 7 )
		);
		write( bytes );
	}
	catch( exception & e )
	{
		cout << e.what() << endl;
	}
}

// TODO implement this
void MackieControlProtocol::notify_name_changed( void *, ARDOUR::Route * route )
{
	try
	{
		cout << "TODO implement MackieControlProtocol::notify_name_changed " << strip_from_route( route ).name();
	}
	catch( exception & e )
	{
		cout << e.what() << endl;
	}
}

// TODO deal with > 1 channel being panned
void MackieControlProtocol::notify_panner_changed( ARDOUR::Route * route )
{
	try
	{
		Pot & pot = strip_from_route( route ).vpot();
		float pos;
		route->panner()[0]->get_effective_position( pos);
		cout << "pos: " << pos << endl;
		
		MidiByteArray bytes( 3
			, midi_pot_id
			// 0x30 + pot number (0-7)
			, 0x30 + pot.ordinal() - 1 
			// the value
			, calculate_pot_value( midi_pot_mode_dot, pos )
		);
		write( bytes );
	}
	catch( exception & e )
	{
		cout << e.what() << endl;
	}
}
