/*
	Copyright (C) 2006,2007 John Anderson

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef ardour_mackie_control_protocol_h
#define ardour_mackie_control_protocol_h

#include <vector>

#include <sys/time.h>
#include <pthread.h>
#include <usb.h>

#include <glibmm/thread.h>

#include <ardour/types.h>
#include <midi++/types.h>

#include <control_protocol/control_protocol.h>
#include "midi_byte_array.h"
#include "controls.h"
#include "route_signal.h"

namespace MIDI {
	class Port;
	class Parser;
}

namespace Mackie {
	class Control;
	class Strip;
}

/**
	This handles the plugin duties, and the midi encoding and decoding,
	and the signal callbacks, mostly from ARDOUR::Route.

	The model of the control surface is handled by classes in controls.h

	What happens is that each strip on the control surface has
	a corresponding route in ControlProtocol::route_table. When
	an incoming midi message is signaled, the correct route
	is looked up, and the relevant changes made to it.

	For each route currently in route_table, there's a RouteSignal object
	which encapsulates the signals that indicate that there are changes
	to be sent to the surface. The signals are handled by this class.

	Calls to signal handlers pass a Route object which is used to look
	up the relevant Strip in Surface. Then the state is retrieved from
	the Route and encoded as the correct midi message.
*/
class MackieControlProtocol : public ARDOUR::ControlProtocol
{
  public:
	MackieControlProtocol (ARDOUR::Session&);
	virtual ~MackieControlProtocol();

	int set_active (bool yn);

	XMLNode& get_state ();
	int set_state (const XMLNode&);
  
	static bool probe() { return true; }
	
	/// Handle all control messags
	void handle_midi_any( MIDI::Parser &, MIDI::byte *, size_t );
	
	/// Handle device initialisation
	void handle_midi_sysex( MIDI::Parser &, MIDI::byte *, size_t );
	
	/// Signal handler for Route::solo
	void notify_solo_changed( void *, ARDOUR::Route * );
	/// Signal handler for Route::mute
	void notify_mute_changed( void *, ARDOUR::Route * );
	/// Signal handler for Route::record_enable_changed
	void notify_record_enable_changed( void *, ARDOUR::Route * );
	/// Signal handler for Route::gain_changed ( from IO )
	void notify_gain_changed( void *, ARDOUR::Route * );
	/// Signal handler for Route::name_change
	void notify_name_changed( void *, ARDOUR::Route * );
	/// Signal handler from Panner::Change
	void notify_panner_changed( ARDOUR::Route * );

  protected:
	void switch_banks( uint32_t initial );
	
	MIDI::Port & port () const { return *_port; }
	
	typedef std::vector<Mackie::RouteSignal*> RouteSignals;
	RouteSignals route_signals;
	
	/**
		Return an instance of Control given a type (from enum midi_types)
		and the rest of the midi message.
	*/
	Mackie::Control & lookup_control( int midi_type, const MidiByteArray & bytes );

	Mackie::Surface & surface()
   {
		return _surface;
	}
  
	/// not used - Parser gets all midi messages
	MidiByteArray read( int count_expecting = -1 );
	
	/// an easier way to output bytes via midi
	void write( MidiByteArray & );
  
	/// To handle the device initialisation sysex messages
	enum sysex_state_t
	{
		expecting_host_connection_query
		, host_connection_confirmation
		, initialisation_error
	};
	sysex_state_t sysex_state;
	static MidiByteArray mackie_sysex_hdr;
	
	/**
		Handle a button press for the control and return whether
		the corresponding light should be on or off.
	*/
	bool handle_button( Mackie::Control & control );
	
	/// Not used currently. May be useful later.
	static void* _monitor_work (void* arg);
	void* monitor_work ();
	
	/**
		Remote ids are from 0..n (n being number of routes in the session.
		indexes are from 0..k, where k is the number of strips on the
		control surface. Usually 8 or 16.
	
		This translates from remote_id to strip, using _current_initial_bank
	*/
	uint32_t index_from_remote_id( uint32_t remote_id );
	
	/**
		Convenience method used by the notify_XXX methods to find
		a strip corresponding to the Route passed in.
	*/
	Mackie::Strip & strip_from_route( ARDOUR::Route * route );

  private:
  
	/// The Midi port for the surface. There will eventually be a set of these.
	MIDI::Port* _port;
  
	/// No used currently.
	pthread_t       thread;
  
	/// The initial remote_id of the currently switched in bank.
   uint32_t _current_initial_bank;
	
	//~ Glib::Mutex update_lock;
	//~ Glib::Mutex io_lock;

	int open ();
	int close ();

	//~ int process (uint8_t *);
	//~ int update_state();
  
   /// The representation of the physical controls on the surface.
  	Mackie::Surface _surface;
};

#endif // ardour_mackie_control_protocol_h
